#
# Copyright (C) 2014-2016  UAVCAN Development Team  <uavcan.org>
#
# This software is distributed under the terms of the MIT License.
#
# Author: Ben Dyer <ben_dyer@mac.com>
#         Pavel Kirienko <pavel.kirienko@zubax.com>
#

import time
import sys
from logging import getLogger
from .. import UAVCANException


logger = getLogger(__name__)


class DriverError(UAVCANException):
    pass


class TxQueueFullError(DriverError):
    pass


class CANFrame:

    def __init__(self, can_id, data, extended, ts_monotonic=None, ts_real=None, canfd=False):
        self.id = can_id
        self.data = data
        self.extended = extended
        self.ts_monotonic = ts_monotonic or time.monotonic()
        self.ts_real = ts_real or time.time()
        self.canfd = canfd
        if self.canfd:
            self.MAX_DATA_LENGTH = 64
        else:
            self.MAX_DATA_LENGTH = 8
        if isinstance(self.data, list):
            self.data = bytearray(self.data)

    def __str__(self):
        if sys.version_info[0] > 2:
            b2int = lambda x: x
        else:
            b2int = ord

        id_str = ('%s %0*x' % ('FD' if self.canfd else '',8 if self.extended else 3, self.id)).rjust(11)
        hex_data = ' '.join(['%02x' % b2int(x) for x in self.data]).ljust(3 * self.MAX_DATA_LENGTH)
        ascii_data = ''.join([(chr(x) if 32 <= x <= 126 else '.') for x in self.data])

        return "%12.6f %12.6f  %s  %s  '%s'" % \
               (self.ts_monotonic, self.ts_real, id_str, hex_data, ascii_data)

    def dlc_to_datalength(dlc):
        # Data Length Code      9  10  11  12  13  14  15
        # Number of data bytes 12  16  20  24  32  48  64
        if (dlc <= 8):
            return dlc
        elif (dlc == 9):
            return 12
        elif (dlc == 10):
            return 16
        elif (dlc == 11):
            return 20
        elif (dlc == 12):
            return 24
        elif (dlc == 13):
            return 32
        elif (dlc == 14):
            return 48
        return 64

    def datalength_to_dlc(data_length):
        if (data_length <= 8):
            return data_length
        elif (data_length <= 12):
            return 9
        elif (data_length <= 16):
            return 10
        elif (data_length <= 20):
            return 11
        elif (data_length <= 24):
            return 12
        elif (data_length <= 32):
            return 13
        elif (data_length <= 48):
            return 14
        return 15

    __repr__ = __str__


class AbstractDriver(object):
    FRAME_DIRECTION_INCOMING = 'rx'
    FRAME_DIRECTION_OUTGOING = 'tx'

    class HookRemover:
        def __init__(self, remover):
            self.remove = remover

    def __init__(self):
        self._io_hooks = []

    def add_io_hook(self, hook):
        """
        Args:
            hook:   This hook will be invoked for every incoming and outgoing CAN frame.
                    Hook arguments: (direction, frame)
                    See FRAME_DIRECTION_*, CANFrame.
        """
        def proxy(*args):
            hook(*args)

        self._io_hooks.append(proxy)

        return self.HookRemover(lambda: self._io_hooks.remove(proxy))

    def _call_io_hooks(self, direction, frame):
        for h in self._io_hooks:
            try:
                h(direction, frame)
            except Exception as ex:
                logger.error('Uncaught exception from CAN IO hook: %r', ex, exc_info=True)

    def _tx_hook(self, frame):
        self._call_io_hooks(self.FRAME_DIRECTION_OUTGOING, frame)

    def _rx_hook(self, frame):
        self._call_io_hooks(self.FRAME_DIRECTION_INCOMING, frame)

    def set_filter_list(self, ids):
        '''set list of message IDs to accept, sent to the remote capture node with mavcan'''
        pass

    def get_filter_list(self, ids):
        '''get list of message IDs to accept, None means accept all'''
        return None
    
    def set_bus(self, busnum):
        '''set the remote bus number to attach to'''
        pass

    def get_bus(self):
        '''get the remote bus number we are attached to'''
        return None

    def get_filter_list(self):
        '''get the current filter list'''
        return None
    
    def set_bus(self, busnum):
        '''set the remote bus number to attach to'''
        pass

    def set_signing_passphrase(self, passphrase):
        '''set MAVLink2 signing passphrase'''
        pass

    def send(self, message_id, message, extended=False, canfd=False):
        self.send_frame(CANFrame(message_id, message, extended, canfd=canfd))



#
# Copyright (C) 2022 DroneCAN Development Team  <uavcan.org>
#
# This software is distributed under the terms of the MIT License.
#
'''
 driver for CAN over MAVLink, using MAV_CMD_CAN_FORWARD and CAN_FRAME messages
'''

import os
import sys
import time
import multiprocessing
from logging import getLogger
from .common import DriverError, CANFrame, AbstractDriver
from pymavlink import mavutil

try:
    import queue
except ImportError:
    # noinspection PyPep8Naming,PyUnresolvedReferences
    import Queue as queue

if 'darwin' in sys.platform:
    RX_QUEUE_SIZE = 32767   # http://stackoverflow.com/questions/5900985/multiprocessing-queue-maxsize-limit-is-32767
else:
    RX_QUEUE_SIZE = 1000000
TX_QUEUE_SIZE = 1000

logger = getLogger(__name__)

class ControlMessage(object):
    def __init__(self, command, data):
        self.command = command
        self.data = data

def io_process(url, bus, target_system, baudrate, tx_queue, rx_queue):
    os.environ['MAVLINK20'] = '1'

    target_component = 0
    last_enable = time.time()
    conn = None
    filter_list = None
    signing_key = None
    last_loss_print_t = time.time()

    def connect():
        nonlocal conn, baudrate
        conn = mavutil.mavlink_connection(url, baud=baudrate, source_system=250,
                                          source_component=mavutil.mavlink.MAV_COMP_ID_MAVCAN,
                                          dialect='ardupilotmega')
        if conn is None:
            raise DriverError('unable to connect to %s' % url)
        nonlocal signing_key
        if signing_key is not None:
            conn.setup_signing(signing_key, sign_outgoing=True)

    def reconnect():
        while True:
            try:
                time.sleep(1)
                logger.info('reconnecting to %s' % url)
                connect()
                return
            except Exception:
                continue

    def enable_can_forward():
        '''re-enable CAN forwarding. Called at 1Hz'''
        nonlocal last_enable, bus, target_system
        last_enable = time.time()
        conn.mav.command_long_send(
            target_system,
            target_component,
            mavutil.mavlink.MAV_CMD_CAN_FORWARD,
            0,
            bus+1,
            0,
            0,
            0,
            0,
            0,
            0)
        if filter_list is not None:
            ids = sorted(filter_list[:16])
            num_ids = len(ids)
            if len(ids) < 16:
                ids += [0]*(16-num_ids)
            try:
                conn.mav.can_filter_modify_send(
                    target_system,
                    target_component,
                    bus+1,
                    mavutil.mavlink.CAN_FILTER_REPLACE,
                    num_ids,
                    ids)
            except Exception as ex:
                print(ex)

    def handle_control_message(m):
        '''handle a ControlMessage'''
        if m.command == "BusNum":
            nonlocal bus
            bus = int(m.data)
        elif m.command == "FilterList":
            nonlocal filter_list
            filter_list = m.data
        elif m.command == "SigningKey":
            nonlocal signing_key
            signing_key = m.data
            conn.setup_signing(signing_key, sign_outgoing=True)

    connect()
    enable_can_forward()

    while True:
        while not tx_queue.empty():
            frame = tx_queue.get()
            if isinstance(frame, ControlMessage):
                handle_control_message(frame)
                continue
            message_id = frame.id
            if frame.extended:
                message_id |= 1<<31
            message = frame.data
            mlen = len(message)
            if mlen < frame.MAX_DATA_LENGTH:
                message += bytearray([0]*(frame.MAX_DATA_LENGTH-mlen))
            try:
                if frame.canfd:
                    conn.mav.canfd_frame_send(
                        target_system,
                        target_component,
                        bus,
                        mlen,
                        message_id,
                        message)
                else:
                    conn.mav.can_frame_send(
                        target_system,
                        target_component,
                        bus,
                        mlen,
                        message_id,
                        message)
            except Exception as ex:
                print(ex)
            if time.time() - last_enable > 1:
                enable_can_forward()

        try:
            m = conn.recv_match(type=['CAN_FRAME','CANFD_FRAME'],blocking=True,timeout=0.005)
        except Exception as ex:
            reconnect()
            continue
        now = time.time()
        if m is None:
            if now - last_enable > 1:
                enable_can_forward()
            if now - last_loss_print_t > 5:
                last_loss_print_t = now
                print("MAVLink packet loss %.2f%%" % conn.packet_loss())
            continue
        if target_system == 0:
            target_system = m.get_srcSystem()
        is_extended = (m.id & (1<<31)) != 0
        is_canfd = m.get_type() == 'CANFD_FRAME'
        canid = m.id & 0x1FFFFFFF
        frame = CANFrame(canid, m.data[:m.len], is_extended, canfd=is_canfd)
        rx_queue.put_nowait(frame)


# MAVLink CAN driver
#
class MAVCAN(AbstractDriver):
    """
    Driver for MAVLink CAN bus adapters, using CAN_FRAME MAVLink packets
    """

    def __init__(self, url, **kwargs):
        super(MAVCAN, self).__init__()
        self.bus = kwargs.get('bus_number', 1) - 1
        self.target_system = kwargs.get('mavlink_target_system', 0)
        self.filter_list = None
        baudrate = kwargs.get('baudrate', 115200)

        self.rx_queue = multiprocessing.Queue(maxsize=RX_QUEUE_SIZE)
        self.tx_queue = multiprocessing.Queue(maxsize=TX_QUEUE_SIZE)

        self.proc = multiprocessing.Process(target=io_process, name='mavcan_io_process',
                                            args=(url, self.bus, self.target_system, baudrate,
                                            self.tx_queue, self.rx_queue))
        self.proc.daemon = True
        self.proc.start()

        # allow signing pass phrase in environment
        pass_phrase = os.environ.get("DRONECAN_SIGNING_KEY",None)
        if pass_phrase:
            self.set_signing_passphrase(pass_phrase)

    def close(self):
        pass

    def __del__(self):
        self.close()

    def receive(self, timeout=None):
        tstart = time.time()
        while True:
            try:
                frame = self.rx_queue.get(block=0)
            except queue.Empty:
                frame = None
            if frame is not None:
                self._rx_hook(frame)
                return frame
            if timeout is not None:
                timeout = max(timeout, 0.001)
                if time.time() >= tstart + timeout:
                    return

    def send_frame(self, frame):
        self._tx_hook(frame)
        self.tx_queue.put_nowait(frame)

    def is_mavlink_port(device_name, baudrate):
        '''check if a device is sending mavlink'''
        os.environ['MAVLINK20'] = '1'
        conn = mavutil.mavlink_connection(device_name, baud=baudrate, source_system=250, source_component=mavutil.mavlink.MAV_COMP_ID_MAVCAN)
        if not conn:
            return False
        m = conn.recv_match(blocking=True, type=['HEARTBEAT','ATTITUDE', 'SYS_STATUS'], timeout=1.1)
        conn.close()
        return m is not None

    def set_filter_list(self, ids):
        '''set list of message IDs to accept, sent to the remote capture node with mavcan'''
        self.filter_list = ids
        self.tx_queue.put_nowait(ControlMessage('FilterList', self.filter_list))

    def get_filter_list(self, ids):
        '''set list of message IDs to accept, sent to the remote capture node with mavcan'''
        return self.filter_list

    def set_bus(self, busnum):
        '''set the remote bus number to attach to'''
        if busnum <= 0:
            raise DriverError('invalid bus %s' % busnum)
        self.bus = busnum - 1
        self.tx_queue.put_nowait(ControlMessage('BusNum', self.bus))

    def get_bus(self):
        '''get the remote bus number we are attached to'''
        return self.bus+1

    def get_filter_list(self):
        '''get the current filter list'''
        return self.filter_list

    def passphrase_to_key(self, passphrase):
        '''convert a passphrase to a 32 byte key'''
        import hashlib
        h = hashlib.new('sha256')
        if sys.version_info[0] >= 3:
            passphrase = passphrase.encode('ascii')
        h.update(passphrase)
        return h.digest()

    def set_signing_passphrase(self, passphrase):
        '''set MAVLink2 signing passphrase'''
        signing_key = self.passphrase_to_key(passphrase)
        self.tx_queue.put_nowait(ControlMessage('SigningKey', signing_key))



    



#!/usr/bin/env python3
'''
bridge between two CAN interfaces
this can be used to bridge a hardware device with SLCAN to a SITL vcan0 environment
'''

import dronecan
import time
import sys
import threading

# get command line arguments
from argparse import ArgumentParser
parser = ArgumentParser(description='CAN bridge - connect two CAN interfaces')
parser.add_argument("--bitrate", default=1000000, type=int, help="CAN bit rate")
parser.add_argument("port1", default=None, type=str, help="first interface")
parser.add_argument("port2", default=None, type=str, help="second interface")
args = parser.parse_args()

try:
    d1 = dronecan.driver.make_driver(args.port1, bitrate=args.bitrate)
except Exception as ex:
    print("Unable to connect to %s - %s" % (args.port1, ex))
    sys.exit(1)
print("Connected to %s" % args.port1)

try:
    d2 = dronecan.driver.make_driver(args.port2, bitrate=args.bitrate)
except Exception as ex:
    print("Unable to connect to %s - %s" % (args.port2, ex))
    sys.exit(1)
print("Connected to %s" % args.port2)

class BridgeThread(object):
    def __init__(self, d1, d2, name):
        self.d1 = d1
        self.d2 = d2
        self.count = 0
        self.thd = threading.Thread(target=self.loop, name=name)
        self.thd.start()

    def loop(self):
        while True:
            try:
                frame = self.d1.receive(timeout=0.1)
            except dronecan.driver.common.TransferError:
                continue
            if frame:
                self.count += 1
                try:
                    self.d2.send_frame(frame)
                except dronecan.driver.common.TxQueueFullError:
                    pass

t1 = BridgeThread(d1, d2, "d1->d2")
t2 = BridgeThread(d2, d1, "d2->d1")

last_print = time.time()

last_d1 = 0
last_d2 = 0

while True:
    time.sleep(1)
    now = time.time()
    dt = now - last_print
    c1 = t1.count
    c2 = t2.count
    print("%.3f/%.3f pkts/sec" % ((c1-last_d1)/dt, (c2-last_d2)/dt))
    last_print = now
    last_d1 = c1
    last_d2 = c2



#
# Copyright (C) 2023 DroneCAN Development Team  <dronecan.org>
#
# This software is distributed under the terms of the MIT License.
#
'''
 driver for CAN over UDP multicast

 multicast address is 239.65.82.B, last octet is logical bus number
 UDP port number is 57732

 packet format is

   - MAGIC 0x2934 16 bit
   - 16 bit CRC CRC-16-CCITT (over all bytes after CRC)
   - 16 bit flags
   - 32 bit message ID
   - data[]

 all data is little endian

 standard URLs
   mcast:  -> first bus
   mcast:0  -> first bus
   mcast:1  -> 2nd bus
'''

MCAST_ADDRESS_BASE = "239.65.82"
MCAST_PORT = 57732
MCAST_MAGIC = 0x2934
MCAST_FLAG_CANFD = 0x0001
MCAST_MAX_PKT_LEN = 74 # 64 byte data + 10 byte header

import os
import sys
import time
import multiprocessing
import socket
import struct
import errno
import select
import dronecan.dsdl.common as common
from logging import getLogger
from .common import DriverError, CANFrame, AbstractDriver

try:
    import queue
except ImportError:
    # noinspection PyPep8Naming,PyUnresolvedReferences
    import Queue as queue

if 'darwin' in sys.platform:
    RX_QUEUE_SIZE = 32767   # http://stackoverflow.com/questions/5900985/multiprocessing-queue-maxsize-limit-is-32767
else:
    RX_QUEUE_SIZE = 1000000
TX_QUEUE_SIZE = 1000

logger = getLogger(__name__)

RUNNING_ON_WINDOWS = False
try:
    # noinspection PyUnresolvedReferences
    sys.getwindowsversion()
    RUNNING_ON_WINDOWS = True
except Exception:
    pass

def io_process(url, tx_queue, rx_queue):
    port = None
    port_out = None
    myport = None
    need_connect = True

    def connect():
        nonlocal port, port_out

        a = url.split(':')
        # acceptable URL forms
        # mcast:  -> DEFAULT_MCAST_ADDRESS
        mcast_port = MCAST_PORT
        mcast_bus = 0
        if len(a) == 1 and len(a[0]) > 0:
            mcast_bus = int(a[0])
        mcast_ip = MCAST_ADDRESS_BASE + ".%u" % mcast_bus

        port = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        if port is None:
            raise DriverError('unable to connect to %s' % url)
        port.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if RUNNING_ON_WINDOWS:
            port.bind(('0.0.0.0', mcast_port))
        else:
            port.bind((mcast_ip, mcast_port))
        mreq = struct.pack("4sl", socket.inet_aton(mcast_ip), socket.INADDR_ANY)
        port.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        port.setblocking(0)

        port_out = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if port_out is None:
            raise DriverError('unable to connect to %s' % url)
        port_out.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        port_out.setblocking(0)
        port_out.connect((mcast_ip, mcast_port))


    while True:
        if need_connect:
            time.sleep(0.1)
            need_connect = False
            connect()
            continue

        while not tx_queue.empty():
            frame = tx_queue.get()
            message_id = frame.id
            if frame.extended:
                message_id |= 1<<31
            flags = MCAST_FLAG_CANFD if frame.canfd else 0
            body = struct.pack("<HI", flags, message_id) + frame.data
            hdr = struct.pack("<HH", MCAST_MAGIC, common.crc16_from_bytes(body))
            pkt = hdr + body

            try:
                port_out.send(pkt)
            except Exception as ex:
                need_connect = True
                continue

        try:
            select.select([port.fileno()], [], [], 0.01)
        except Exception as ex:
            need_connect = True
            pass

        try:
            data, new_addr = port.recvfrom(MCAST_MAX_PKT_LEN)
            if myport is None:
                try:
                    (myaddr,myport) = port_out.getsockname()
                except Exception:
                    pass
        except socket.error as e:
            if e.errno in [ errno.EAGAIN, errno.EWOULDBLOCK, errno.ECONNREFUSED ]:
                continue
            need_connect = True
            continue
        if myport == new_addr[1]:
            # data from ourselves, discard
            continue

        if len(data) < 10:
            continue
        magic, crc16, flags, message_id = struct.unpack("<HHHI", data[:10])
        if magic != MCAST_MAGIC:
            continue
        if crc16 != common.crc16_from_bytes(data[4:]):
            continue

        is_extended = (message_id & (1<<31)) != 0
        is_canfd = flags & MCAST_FLAG_CANFD != 0
        canid = message_id & 0x1FFFFFFF
        frame = CANFrame(canid, data[10:], is_extended, canfd=is_canfd)
        rx_queue.put_nowait(frame)


# UDP multicast CAN driver
#
class mcast(AbstractDriver):
    """
    Driver for multicast UDP CAN
    """

    def __init__(self, url, **kwargs):
        super(mcast, self).__init__()

        self.rx_queue = multiprocessing.Queue(maxsize=RX_QUEUE_SIZE)
        self.tx_queue = multiprocessing.Queue(maxsize=TX_QUEUE_SIZE)

        self.proc = multiprocessing.Process(target=io_process, name='mcast_io_process',
                                            args=(url, self.tx_queue, self.rx_queue))
        self.proc.daemon = True
        self.proc.start()

    def close(self):
        pass

    def __del__(self):
        self.close()

    def receive(self, timeout=None):
        tstart = time.time()
        while True:
            try:
                frame = self.rx_queue.get(block=0)
            except queue.Empty:
                frame = None
            if frame is not None:
                self._rx_hook(frame)
                return frame
            if timeout is not None:
                timeout = max(timeout, 0.001)
                if time.time() >= tstart + timeout:
                    return

    def send_frame(self, frame):
        self._tx_hook(frame)
        self.tx_queue.put_nowait(frame)



'''
utility classes for working with DroneCAN
'''

import dronecan
import time

class DroneCANSerial(object):
    '''emulation of a serial port over DroneCAN'''
    def __init__(self, uri, target_node, target_serial_dev, lock_port=False, node=None, baudrate=230400, node_id=100, timeout=2.0):
        """Constructor.

        :param uri: DroneCAN URI
        :param target_node: CAN node number hosting serial port
        :param target_serial_dev: serial device number or -1 for automatic selection
        :param lock_port: lock the serial port while in use
        :param node: DroneCAN node, if None then one is created
        :param baudrate: initial baud rate
        :param node_id: local node id if node not supplied
        :param timeout: timeout of read operations

        """
        self.buf = bytearray()
        self.timeout = timeout
        self.target_node = target_node
        self.target_serial_dev = target_serial_dev
        self.node_id = node_id
        self.lock_port = lock_port
        self.node = node
        if node is None:
            self.node = dronecan.make_node(uri, node_id=self.node_id)
        self.handle = self.node.add_handler(dronecan.uavcan.tunnel.Targetted, self.handle_Targetted)
        self.baudrate = baudrate
        self.last_send = 0

    def spin(self):
        '''spin once to process packets'''
        try:
            while self.node.spin(0) > 0:
                pass
        except dronecan.transport.TransferError:
            # ignore corrupt frames
            pass

    def __del__(self):
        '''close port'''
        self.close()

    def close(self):
        '''close port'''
        self.handle.remove()
        self.node = None

    def handle_Targetted(self, msg):
        if msg.transfer.source_node_id != self.target_node:
            return
        if msg.message.target_node != self.node.node_id:
            return
        self.buf.extend(msg.message.buffer)

    def send_bytes(self, b):
        '''send some bytes'''
        n = len(b)
        if n > 120:
            n = 120
        message = dronecan.uavcan.tunnel.Targetted()
        message.protocol.protocol = 2
        message.target_node = self.target_node
        message.serial_id = self.target_serial_dev
        message.options = message.OPTION_LOCK_PORT if self.lock_port else 0
        message.baudrate = self.baudrate
        message.buffer = bytearray(b[:n])
        while True:
            # implement writes as blocking
            try:
                self.spin()
                self.node.broadcast(message)
            except dronecan.driver.common.TxQueueFullError:
                time.sleep(0.01)
                continue
            break
        self.last_send = time.time()
        return n

    def write(self, b):
        '''write some bytes'''
        total = 0
        while len(b) > 0:
            n = self.send_bytes(b)
            if n <= 0:
                break
            total += n
            b = b[n:]
        return total

    def read(self, n):
        '''read some bytes'''
        # ensure we have recent packets
        self.spin()
        if time.time() - self.last_send > 0.5:
            # keep connection alive
            self.send_bytes(bytearray())
        if len(self.buf) > 0:
            if n > len(self.buf):
                n = len(self.buf)
            ret = self.buf[:n]
            self.buf = self.buf[n:]
            return ret
        return bytearray()

    def flushInput(self):
        '''flush any pending input'''
        self.buf = bytearray()

    def setBaudrate(self, baudrate):
        '''set baudrate'''
        if self.baudrate == baudrate:
            return
        self.baudrate = baudrate
        self.flushInput()

    def flush(self):
        '''ensure data is written'''
        self.spin()
